package gg.malloc.defense.engine;

import gg.malloc.defense.model.Arena;
import gg.malloc.defense.model.Game;
import gg.malloc.defense.model.Spawner;
import gg.malloc.defense.model.Waypoint;

import net.md_5.bungee.api.chat.BaseComponent;
import net.md_5.bungee.api.chat.ClickEvent;
import net.md_5.bungee.api.ChatColor;
import net.md_5.bungee.api.chat.ComponentBuilder;

import gg.malloc.defense.ui.BombCarrier;
import gg.malloc.defense.ui.BossBars;
import gg.malloc.defense.ui.Items;

import gg.malloc.defense.Plugin;

import java.util.logging.Logger;
import java.util.ArrayList;
import java.util.HashSet;

import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Particle;
import org.bukkit.Sound;
import org.bukkit.SoundCategory;
import org.bukkit.World;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityTargetEvent;
import org.bukkit.event.player.PlayerTeleportEvent;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.Material;
import org.bukkit.attribute.Attribute;
import org.bukkit.attribute.AttributeModifier;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;

public class GameRunner {
  Arena m_arena;
  Game m_game;
  Stage m_stage;
  Plugin m_plugin;

  MobManager m_mobs;
  WaveManager m_waves;
  PlayerManager m_players;

  BombFuse m_bombFuse;

  Logger m_log;

  public enum Stage {
    Idle,
    Warmup,
    Countdown,
    Playing,
    GameOver
  }

  TickTask m_fuseTask;
  TickTask m_countdownTask;
  TickTask m_lobbyReturnTask;
  TickTask m_bombSmokeTask;
  TickTask m_bombCrackleTask;

  BossBars m_bars;

  World m_world;

  Location getLocation(Waypoint waypoint) {
    return new Location(m_world, waypoint.getX(), waypoint.getY(), waypoint.getZ());
  }

  public GameRunner(Plugin plugin, Game game, Arena arena, World world) {
    m_plugin = plugin;
    m_game = game;
    m_world = world;
    m_arena = arena;
    m_stage = Stage.Idle;
    m_mobs = new MobManager();
    m_waves = new WaveManager(m_game);
    m_players = new PlayerManager();
    m_bombFuse = new BombFuse();
    m_bars = new BossBars(this, m_game, m_waves, m_players, m_bombFuse, m_mobs);
    m_log = m_plugin.getLogger();

    m_fuseTask = new TickTask(m_plugin, () -> {
      if (m_bombFuse.isLit()) {
        m_bombFuse.tickDecay();
        m_bars.update();
      }
    }, 80);

    m_countdownTask = new TickTask(m_plugin, () -> {
      if (m_warmupCountdown == 0) {
        requestTransition(Stage.Playing);
      } else {
        BaseComponent[] message = new ComponentBuilder()
          .append("Starting game in ").color(ChatColor.LIGHT_PURPLE)
          .append("" + m_warmupCountdown).color(ChatColor.AQUA).bold(true)
          .create();
        broadcastMessage(message);
        m_warmupCountdown--;
        m_bars.setCountdownProgress((double)m_warmupCountdown / 10.0);
        m_bars.update();
      }
    }, 20);

    m_lobbyReturnTask = new TickTask(m_plugin, () -> {
      if (m_gameEndCountdown == 0) {
        for(Player p : new ArrayList<Player>(m_players.getPlayers())) {
          removePlayer(p);
          m_plugin.returnPlayerToLobby(p);
        }
        requestTransition(Stage.Idle);
      } else {
        BaseComponent[] message = new ComponentBuilder()
          .append("Game is ending in ").color(ChatColor.LIGHT_PURPLE)
          .append("" + m_gameEndCountdown).color(ChatColor.RED).bold(true)
          .create();
        broadcastMessage(message);
        m_gameEndCountdown--;
        m_bars.setCountdownProgress((double)m_gameEndCountdown / 30.0);
        m_bars.update();
      }
    }, 20);

    m_bombSmokeTask = new TickTask(m_plugin, () -> {
      Location targetLoc = getLocation(m_arena.bombTarget());
      m_world.spawnParticle(Particle.SMOKE_LARGE, targetLoc, 35, 4, 2, 4);
      m_world.spawnParticle(Particle.SMALL_FLAME, targetLoc, 30, 3, 2, 3);
    }, 5);
    m_bombCrackleTask = new TickTask(m_plugin, () -> {
      Location targetLoc = getLocation(m_arena.bombTarget());
      m_world.playSound(targetLoc, Sound.BLOCK_CAMPFIRE_CRACKLE, SoundCategory.NEUTRAL, 1.0f, 1.0f);
    }, 35);
  }

  int m_warmupCountdown = 0;
  int m_gameEndCountdown = 0;

  public void handleEntityRetargeting(EntityTargetEvent evt) {
    m_mobs.handleEntityRetarget(evt);
  }

  public void handleEntityDamage(EntityDamageEvent evt) {
    m_mobs.handleEntityDamage(evt);
    if (m_mobs.bombWasHit()) {
      EntityDamageByEntityEvent entityEvt = (EntityDamageByEntityEvent)evt;
      m_log.info("Target attacked!");
      entityEvt.getDamager().setGlowing(true);
      m_bombFuse.tickLit();
      m_world.playSound(getLocation(m_arena.bombTarget()), Sound.ENTITY_ZOMBIE_ATTACK_IRON_DOOR, 1.5f, 0.9f);
      m_bars.update();
      if (m_bombFuse.isExploded()) {
        m_world.strikeLightningEffect(getLocation(m_arena.bombTarget()));
        m_world.playSound(getLocation(m_arena.bombTarget()), Sound.ENTITY_GENERIC_EXPLODE, SoundCategory.NEUTRAL, 1.3f, 1.0f);
        m_world.spawnParticle(Particle.EXPLOSION_HUGE, getLocation(m_arena.bombTarget()), 8, 5, 2, 5);
        BaseComponent[] message = new ComponentBuilder()
          .append("The bomb blew up!").color(ChatColor.RED).bold(true)
          .create();
        broadcastMessage(message);
        requestTransition(Stage.GameOver);
        m_bombSmokeTask.start();
        m_bombCrackleTask.start();
      }
    } else if (evt.getEntity() instanceof Player) {
      Player player = (Player)evt.getEntity();
      if (evt instanceof EntityDamageByEntityEvent) {
        EntityDamageByEntityEvent entityEvt = (EntityDamageByEntityEvent)evt;
        if (entityEvt.getDamager() instanceof Player) {
          evt.setCancelled(true);
          return;
        }
      }
      if (m_players.contains(player) && player.getHealth() - evt.getFinalDamage() <= 0) {
        evt.setCancelled(true);
        handlePlayerDeath(player);
      }
    }
  }

  private boolean enterIdle() {
    m_bombFuse.reset();
    m_waves.reset();
    m_mobs.clear();
    m_bombSmokeTask.stop();
    m_bombCrackleTask.stop();
    m_players.requestTransitionForAll(PlayerManager.State.Idle);
    m_fuseTask.stop();
    m_countdownTask.stop();
    m_lobbyReturnTask.stop();
    return true;
  }

  private boolean enterWarmup() {
    m_waves.next();
    for(Player p : m_players.getPlayers()) {
        if (m_players.requestTransition(p, PlayerManager.State.Playing)) {
          p.teleport(m_world.getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
        }
        m_players.setReady(p, false);
    }
    BaseComponent[] message = new ComponentBuilder()
      .append("Click").color(ChatColor.LIGHT_PURPLE)
      .append("[Here]").color(ChatColor.GOLD)
      .event(new ClickEvent(ClickEvent.Action.RUN_COMMAND, "/malloc-defense:ready"))
      .append(" when ready.").color(ChatColor.LIGHT_PURPLE)
      .create();
    broadcastMessage(message);
    m_mobs.clear();
    return true;
  }

  public void togglePlayerReady(Player p) {
    setPlayerReady(p, !m_players.isReady(p));
  }

  void clearReadyState() {
    for(Player p : m_players.getPlayers()) {
      setPlayerReady(p, false);
    }
  }

  public void setPlayerReady(Player p, boolean isReady) {
    if (isReady != m_players.isReady(p)) {
      if (isReady) {
        BaseComponent[] message = new ComponentBuilder()
          .append(p.getName()).bold(true)
          .append(" is ready!").color(ChatColor.AQUA).italic(true)
          .create();
      }
      m_players.setReady(p, isReady);
    }
    m_bars.update();
    broadcastMessage(message);
    if (m_players.isEveryoneReady()) {
      requestTransition(Stage.Countdown);
    }
  }

  private boolean enterCountdown() {
    m_warmupCountdown = 10;
    m_bars.setCountdownProgress(1.0);
    m_countdownTask.start();
    return true;
  }

  private boolean enterPlaying() {
    m_log.info("Starting wave " + m_waves.currentWaveNum());
    m_mobs.spawnTarget(getLocation(m_arena.bombTarget()));
    // TODO: Set helmet with custom model data
    for(Player p : m_players.getPlayers()) {
        if (m_players.requestTransition(p, PlayerManager.State.Playing)) {
          p.teleport(m_world.getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
        }
    }
    m_countdownTask.stop();
    m_fuseTask.start();
    spawnNextBatch();
    return true;
  }
  
  private boolean enterGameOver() {
    m_gameEndCountdown = 60;
    m_lobbyReturnTask.start();
    m_countdownTask.stop();
    m_fuseTask.stop();
    m_mobs.clear();
    return true;
  }

  public Stage getStage() {
    return m_stage;
  }

  private void spawnNextBatch() {
    m_log.info("Spawning batch " + m_waves.currentBatchNum());
    Spawner spawner = new GameSpawner(m_world, m_arena, m_mobs, m_players);
    m_waves.currentWave().spawnBatch(spawner, m_waves.currentBatchNum());
    m_bars.update();
  }

  private void handlePlayerDeath(Player player) {
    if (m_players.requestTransition(player, PlayerManager.State.Dead)) {
      m_world.strikeLightningEffect(player.getLocation());
      if (!m_players.isAnyoneAlive()) {
        BaseComponent[] message = new ComponentBuilder()
          .append("Everyone is ").color(ChatColor.LIGHT_PURPLE)
          .append("DEAD").color(ChatColor.RED).bold(true)
          .append(" :(").color(ChatColor.LIGHT_PURPLE)
          .create();
        broadcastMessage(message);
        requestTransition(Stage.GameOver);
      } else {
        m_log.info("Remaining players " + m_players.remainingPlayers());
      }
    }
  }

  public void handleEntityDeath(Entity entity) {
    boolean wasCarrier = m_mobs.isBombCarrier(entity);
    if (m_mobs.killMob(entity)) {
      int coinsToDrop = 300;
      while(coinsToDrop > 0) {
        int droppedCoins = Math.min(coinsToDrop, 64);
        ItemStack coins = Items.makeCoins();
        coins.setAmount(droppedCoins);
        coinsToDrop -= 64;
        m_world.dropItem(entity.getLocation(), coins);
      }
      m_bars.update();
      if (m_mobs.remainingMobs() <= 3) {
        m_log.info("Starting next batch!");
        if (m_waves.isLastBatch()) {
          if (m_waves.isLastWave()) {
            BaseComponent[] message = new ComponentBuilder()
              .append("You Won!").color(ChatColor.LIGHT_PURPLE).bold(true)
              .create();
            broadcastMessage(message);
            requestTransition(Stage.GameOver);
          } else if (m_mobs.empty()) {
            requestTransition(Stage.Warmup);
          }
        } else {
          m_waves.nextBatch();
          spawnNextBatch();
        }
      } else {
        m_log.fine("Living mobs remaining: " + m_mobs.remainingMobs());
      }
    }
  }

  private boolean attemptTransition(Stage stage) {
    if (m_stage == stage) {
      return false;
    }
    switch(stage) {
      case Idle:
        return enterIdle();
      case Warmup:
        return enterWarmup();
      case Countdown:
        return enterCountdown();
      case Playing:
        return enterPlaying();
      case GameOver:
        return enterGameOver();
    }
    return false;
  }
  
  private boolean validateTransition(Stage from, Stage to) {
    switch(from) {
      case Idle:
        return !m_players.isEmpty() && to == Stage.Warmup;
      case Warmup:
        return to == Stage.Playing || to == Stage.Idle || to == Stage.Countdown;
      case Countdown:
        return to == Stage.Playing || to == Stage.Idle || to == Stage.Warmup;
      case Playing:
        return to == Stage.Warmup || to == Stage.GameOver || to == Stage.Idle;
      case GameOver:
        return to == Stage.Idle;
    }
    return false;
  }

  public boolean requestTransition(Stage stage) {
    if (!validateTransition(m_stage, stage)) {
      m_log.warning("Attemped illegal transition: " + m_stage + " -> " + stage);
      return false;
    }
    if (attemptTransition(stage)) {
      m_log.info("Game transition: " + m_stage + " -> " + stage);
      m_stage = stage;
      m_bars.update();
      for(Player p : m_players.getPlayers()) {
        sendStageTitle(p);
      }
      return true;
    }
    m_log.severe("Failed to complete transition: " + m_stage + " -> " + stage);
    return false;
  }

  void sendStageTitle(Player p) {
    switch(m_stage) {
      case Warmup:
        p.sendTitle("Warmup", "Prepare yourself for wave " + m_waves.currentWaveNum(), 10, 70, 20);
        break;
      case Playing:
        p.sendTitle("Wave " + m_waves.currentWaveNum(), "", 10, 70, 20);
        break;
      case GameOver:
        p.sendTitle("Game over!", "", 10, 70, 20);
    }
  }

  public void addPlayer(Player p) {
    if (m_players.contains(p)) {
      return;
    }
    m_players.addPlayer(p);
    m_bars.addPlayer(p);
    if (m_stage == Stage.Idle || m_stage == Stage.Warmup) {
      if (m_players.requestTransition(p, PlayerManager.State.Playing)) {
        p.teleport(m_world.getSpawnLocation(), PlayerTeleportEvent.TeleportCause.PLUGIN);
      }
      BaseComponent[] message = new ComponentBuilder()
        .append(p.getName()).bold(true)
        .append("Has joined the game").color(ChatColor.AQUA)
        .create();
      broadcastMessage(message);
      sendStageTitle(p);
      if (m_stage == Stage.Idle) {
        requestTransition(Stage.Warmup);
      }
    }
  }

  public void removePlayer(Player p) {
    m_bars.removePlayer(p);
    m_players.removePlayer(p);
    if (m_players.isEmpty()) {
      requestTransition(Stage.Idle);
    } else {
      BaseComponent[] message = new ComponentBuilder()
        .append(p.getName()).bold(true)
        .append("Has left the game").color(ChatColor.AQUA)
        .create();
      broadcastMessage(message);
    }
  }

  void broadcastTitle(String title) {
    broadcastTitle(title, "");
  }

  void broadcastTitle(String title, String subtitle) {
    for(Player p : m_players.getPlayers()) {
      p.sendTitle(title, subtitle, 10, 70, 20);
    }
  }

  void broadcastMessage(BaseComponent[] component) {
    for(Player p : m_world.getPlayers()) {
      p.spigot().sendMessage(component);
    }
  }
}
