package gg.malloc.defense.engine;

import java.util.HashSet;

import org.bukkit.Location;
import org.bukkit.entity.ArmorStand;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Mob;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityTargetEvent;

import gg.malloc.defense.model.Game;

import gg.malloc.defense.ui.BombCarrier;

public class MobManager {
  HashSet<LivingEntity> m_livingMobs = new HashSet<>();
  HashSet<LivingEntity> m_bombCarriers = new HashSet<>();
  HashSet<LivingEntity> m_droppedBombs = new HashSet<>();

  int m_createdMobs = 0;
  int m_killedMobs = 0;

  public boolean contains(Entity entity) {
    return m_livingMobs.contains(entity);
  }

  public boolean isBombCarrier(Entity entity) {
    return m_bombCarriers.contains(entity);
  }

  public boolean killMob(Entity entity) {
    if (m_livingMobs.contains(entity)) {
      m_killedMobs += 1;
      if (m_bombCarriers.remove(entity)) {
        dropBomb(entity.getLocation());
      }
      return m_livingMobs.remove(entity);
    }
    return false;
  }

  public void clear() {
    for(LivingEntity e : m_livingMobs) {
      e.remove();
    }
    for(LivingEntity e : m_droppedBombs) {
      e.remove();
    }
    if (m_bombTarget != null) {
      m_bombTarget.remove();
    }
    m_bombTarget = null;
    m_droppedBombs.clear();
    m_livingMobs.clear();
    m_bombCarriers.clear();
    m_createdMobs = 0;
    m_killedMobs = 0;
  }

  public int createdMobs() {
    return m_createdMobs;
  }

  public int killedMobs() {
    return m_killedMobs;
  }

  public int remainingMobs() {
    return m_createdMobs - m_killedMobs;
  }

  public double progress() {
    return (double)m_killedMobs / (double)m_createdMobs;
  }

  public void addEntity(LivingEntity entity) {
    m_livingMobs.add(entity);
    Mob asMob = (Mob)entity;
    if (m_droppedBombs.size() > 0) {
      asMob.setTarget(m_droppedBombs.iterator().next());
    } else {
      asMob.setTarget(m_bombTarget);
    }
    m_createdMobs += 1;
  }

  public int bombCount() {
    return m_bombCarriers.size() + m_droppedBombs.size();
  }

  public LivingEntity addBombCarrier(LivingEntity entity) {
    m_bombCarriers.add(entity);
    ((Mob)entity).setTarget(m_bombTarget);
    return new BombCarrier(entity).inHand();
  }

  public void removeBombCarrier(LivingEntity entity) {
    m_bombCarriers.remove(entity);
  }

  public boolean empty() {
    return m_livingMobs.size() == 0;
  }

  public LivingEntity dropBomb(Location location) {
    ArmorStand droppedBomb = (ArmorStand)(new BombCarrier((LivingEntity)location.getWorld().spawnEntity(location, EntityType.ARMOR_STAND)).onHead());
    droppedBomb.setCustomName("The Bomb");
    droppedBomb.setVisible(true);
    droppedBomb.setCustomNameVisible(true);
    droppedBomb.setGlowing(true);
    m_droppedBombs.add(droppedBomb);
    for(LivingEntity ent : m_livingMobs) {
      if (!m_bombCarriers.contains(ent)) {
        ((Mob)ent).setTarget(droppedBomb);
      }
    }
    return droppedBomb;
  }

  boolean m_bombWasHit = false;

  public boolean bombWasHit() {
    return m_bombWasHit;
  }

  public void handleEntityDamage(EntityDamageEvent evt) {
    Entity entity = evt.getEntity();
    m_bombWasHit = false;

    if (entity == m_bombTarget && evt instanceof EntityDamageByEntityEvent) {
      EntityDamageByEntityEvent entityEvt = (EntityDamageByEntityEvent)evt;
      evt.setCancelled(true);
      if (m_bombCarriers.contains(entityEvt.getDamager())) {
        m_bombWasHit = true;
      }
    } if (m_droppedBombs.contains(entity)) {
      evt.setCancelled(true);
      EntityDamageByEntityEvent entityEvt = (EntityDamageByEntityEvent)evt;
      if (m_livingMobs.contains(entityEvt.getDamager())) {
        evt.getEntity().remove();
        m_droppedBombs.remove(evt.getEntity());
        addBombCarrier((LivingEntity)entityEvt.getDamager());
        for(LivingEntity ent : m_livingMobs) {
          if (!m_bombCarriers.contains(ent)) {
            ((Mob)ent).setTarget(m_bombTarget);
          }
        }
      }
    } else if (m_livingMobs.contains(evt.getEntity())) {
      //m_game.onMobDamaged(evt.getEntity());
    }
  }

  LivingEntity m_bombTarget;

  public LivingEntity spawnTarget(Location location) {
    assert(m_bombTarget == null);
    m_bombTarget = (LivingEntity)location.getWorld().spawnEntity(location, EntityType.ARMOR_STAND);
    ArmorStand bombStand = (ArmorStand)m_bombTarget;
    bombStand.setCustomName("Bomb Target");
    bombStand.setVisible(true);
    bombStand.setCustomNameVisible(true);
    bombStand.setGlowing(true);
    return m_bombTarget;
  }

  public void handleEntityRetarget(EntityTargetEvent evt) {
    if (m_livingMobs.contains(evt.getEntity())) {
      if (m_bombCarriers.contains(evt.getEntity())) {
        evt.setTarget(m_bombTarget);
      } if (m_droppedBombs.size() > 0) {
        evt.setTarget(m_droppedBombs.iterator().next());
      } else if (evt.getReason() == EntityTargetEvent.TargetReason.FORGOT_TARGET) {
        evt.setTarget(m_bombTarget);
      }
    }
  }
}
